import itertools
import logging
import random
from itertools import product

import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LogisticRegression

import darts.metrics as metrics
from darts import TimeSeries
from darts.datasets import AirPassengersDataset, MonthlyMilkDataset
from darts.logging import get_logger
from darts.models import (
    ARIMA,
    FFT,
    ExponentialSmoothing,
    LinearRegressionModel,
    NaiveDrift,
    NaiveSeasonal,
    RandomForestModel,
    SKLearnClassifierModel,
    Theta,
)
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils.timeseries_generation import constant_timeseries as ct
from darts.utils.timeseries_generation import gaussian_timeseries as gt
from darts.utils.timeseries_generation import linear_timeseries as lt
from darts.utils.timeseries_generation import random_walk_timeseries as rt
from darts.utils.timeseries_generation import sine_timeseries as st
from darts.utils.utils import generate_index

logger = get_logger(__name__)


if TORCH_AVAILABLE:
    from darts.models import BlockRNNModel, TCNModel


def get_dummy_series(
    ts_length: int, lt_end_value: int = 10, st_value_offset: int = 10
) -> TimeSeries:
    return (
        lt(length=ts_length, end_value=lt_end_value)
        + st(length=ts_length, value_y_offset=st_value_offset)
        + rt(length=ts_length)
    )


def compare_best_against_random(model_class, params, series, stride=1):
    # instantiate best model in expanding window mode
    np.random.seed(1)
    best_model_1, _, _ = model_class.gridsearch(
        params,
        series,
        forecast_horizon=10,
        stride=stride,
        metric=metrics.mape,
        start=series.time_index[-21],
    )

    # instantiate best model in split mode
    train, val = series.split_before(series.time_index[-10])
    best_model_2, _, _ = model_class.gridsearch(
        params, train, val_series=val, metric=metrics.mape
    )

    # instantiate model with random parameters from 'params'
    random.seed(1)
    random_param_choice = {}
    for key in params.keys():
        random_param_choice[key] = random.choice(params[key])
    random_model = model_class(**random_param_choice)

    # perform backtest forecasting on both models
    best_score_1 = best_model_1.backtest(
        series, start=series.time_index[-21], forecast_horizon=10
    )
    random_score_1 = random_model.backtest(
        series, start=series.time_index[-21], forecast_horizon=10
    )

    # perform train/val evaluation on both models
    best_model_2.fit(train)
    best_score_2 = metrics.mape(best_model_2.predict(len(val)), series)
    random_model = model_class(**random_param_choice)
    random_model.fit(train)
    random_score_2 = metrics.mape(random_model.predict(len(val)), series)

    # check whether best models are at least as good as random models
    expanding_window_ok = best_score_1 <= random_score_1
    split_ok = best_score_2 <= random_score_2

    return expanding_window_ok and split_ok


class TestBacktesting:
    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [False, True],
            [
                ([metrics.mape], 100.0),
                ([metrics.mape, metrics.mape], 100.0),
                ([metrics.f1], 0.0),
                ([metrics.f1, metrics.precision], 0.0),
            ],
        ),
    )
    def test_output_single_series_hfc_lpo_true(self, config):
        """Tests backtest based on historical forecasts generated on a single `series` (or list of one `series`)
        with last_points_only=True"""
        is_univariate, series_as_list, (metric, score_expected) = config
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        y = y if not series_as_list else [y]
        hfc = hfc if not series_as_list else [hfc]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=False,
            )
        if series_as_list:
            error_msg = "Expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]`"
        else:
            error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        if series_as_list:
            with pytest.raises(ValueError) as err:
                _ = model.backtest(
                    series=y,
                    historical_forecasts=hfc + y,
                    reduction=None,
                    metric=metric,
                    last_points_only=True,
                )
            error_msg = f"expected `historical_forecasts` of type `Sequence[TimeSeries]` with length n={len(y)}."
            assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=None,
            metric=metric,
            last_points_only=True,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        if not is_multi_metric:
            # inner type expected: 1 float
            assert isinstance(bt, float) and bt == score_expected
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt, np.ndarray)
            np.testing.assert_array_almost_equal(bt, np.array([score_expected] * 2))

        # with reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=np.mean,
            metric=metric,
            last_points_only=True,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        if not is_multi_metric:
            # inner type expected: 1 float
            assert isinstance(bt, float) and bt == score_expected
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt, np.ndarray)
            np.testing.assert_array_almost_equal(bt, np.array([score_expected] * 2))

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [False, True],
            [
                ([metrics.mape], [0.0, 100.0]),
                ([metrics.mape, metrics.mape], [0.0, 100.0]),
                ([metrics.f1], [1.0, 0.0]),
                ([metrics.f1, metrics.precision], [1.0, 0.0]),
            ],
            [1, 2],
        ),
    )
    def test_output_single_series_hfc_lpo_false(self, config):
        """Tests backtest based on historical forecasts generated on a single `series` (or list of one `series`)
        with last_points_only=False"""
        is_univariate, series_as_list, (metric, score_expected), n_forecasts = config
        score_hi_expected, score_lo_expected = score_expected
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [y, hfc]
        hfc = hfc[:n_forecasts]

        y = y if not series_as_list else [y]
        hfc = hfc if not series_as_list else [hfc]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=True``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        if series_as_list:
            error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        else:
            error_msg = "Expected `historical_forecasts` of type `TimeSeries`"
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        if series_as_list:
            with pytest.raises(ValueError) as err:
                _ = model.backtest(
                    series=y,
                    historical_forecasts=hfc + [y],
                    reduction=None,
                    metric=metric,
                    last_points_only=False,
                )
            error_msg = (
                f"expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]`"
                f" with length n={len(y)}."
            )
            assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=None,
            metric=metric,
            last_points_only=False,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        assert isinstance(bt, np.ndarray)
        if not is_multi_metric:
            # inner shape expected: (n hist forecasts = 2,)
            np.testing.assert_array_almost_equal(
                bt, np.array([score_hi_expected, score_lo_expected])[:n_forecasts]
            )
        else:
            # inner shape expected: (n hist forecasts = 2, n metrics = 2)
            np.testing.assert_array_almost_equal(
                bt,
                np.array([[score_hi_expected] * 2, [score_lo_expected] * 2])[
                    :n_forecasts
                ],
            )

        # with reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=np.mean,
            metric=metric,
            last_points_only=False,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        score_exp = (
            score_hi_expected
            if n_forecasts == 1
            else (score_hi_expected + score_lo_expected) / 2
        )
        if not is_multi_metric:
            # inner shape expected: 1 float
            assert isinstance(bt, float) and bt == score_exp
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt, np.ndarray)
            np.testing.assert_array_almost_equal(bt, np.array([score_exp, score_exp]))

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [
                ([metrics.mape], [0.0, 100.0]),
                ([metrics.mape, metrics.mape], [0.0, 100.0]),
                ([metrics.f1], [1.0, 0.0]),
                ([metrics.f1, metrics.precision], [1.0, 0.0]),
            ],
        ),
    )
    def test_output_multi_series_hfc_lpo_true(self, config):
        """Tests backtest based on historical forecasts generated on multiple `series` with last_points_only=True"""
        is_univariate, (metric, score_expected) = config
        score_hi_expected, score_lo_expected = score_expected
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [y, hfc]
        y = [y, y]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=False,
            )
        error_msg = (
            "Expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]`"
        )
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc + [y[0]],
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        error_msg = f"expected `historical_forecasts` of type `Sequence[TimeSeries]` with length n={len(y)}."
        assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=None,
            last_points_only=True,
            metric=metric,
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # per series, inner type expected: 1 float
            assert bt == score_expected
        else:
            # per series, inner shape expected: (n metrics = 2,)
            assert all(isinstance(bt_, np.ndarray) for bt_ in bt)
            np.testing.assert_array_almost_equal(
                bt[0], np.array([score_hi_expected] * 2)
            )
            np.testing.assert_array_almost_equal(
                bt[1], np.array([score_lo_expected] * 2)
            )

        # with reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=np.mean,
            last_points_only=True,
            metric=metric,
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # per series, inner type expected: 1 float
            assert bt == score_expected
        else:
            # per series, inner shape expected: (n metrics = 2,)
            assert all(isinstance(bt_, np.ndarray) for bt_ in bt)
            np.testing.assert_array_almost_equal(
                bt[0], np.array([score_hi_expected] * 2)
            )
            np.testing.assert_array_almost_equal(
                bt[1], np.array([score_lo_expected] * 2)
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [
                ([metrics.mape], [0.0, 100.0]),
                ([metrics.mape, metrics.mape], [0.0, 100.0]),
                ([metrics.f1], [1.0, 0.0]),
                ([metrics.f1, metrics.precision], [1.0, 0.0]),
            ],
        ),
    )
    def test_output_multi_series_hfc_lpo_false(self, config):
        """Tests backtest based on historical forecasts generated on multiple `series` with
        last_points_only=False.
        """
        is_univariate, (metric, score_expected) = config
        score_hi_expected, score_lo_expected = score_expected
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [[y], [hfc]]
        y = [y, y]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc + [[y[0]]],
                reduction=None,
                metric=metric,
                last_points_only=False,
            )
        error_msg = f"expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]` with length n={len(y)}."
        assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=None, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        assert isinstance(bt[0], np.ndarray)
        assert isinstance(bt[1], np.ndarray)
        if not is_multi_metric:
            # inner shape expected: (n hist forecasts = 1,)
            np.testing.assert_array_almost_equal(bt[0], np.array([score_hi_expected]))
            np.testing.assert_array_almost_equal(bt[1], np.array([score_lo_expected]))
        else:
            # inner shape expected: (n metrics = 2, n hist forecasts = 1)
            np.testing.assert_array_almost_equal(
                bt[0], np.array([[score_hi_expected] * 2])
            )
            np.testing.assert_array_almost_equal(
                bt[1], np.array([[score_lo_expected] * 2])
            )

        # with reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=np.mean, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # inner type expected: 1 float
            assert bt == score_expected
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt[0], np.ndarray)
            np.testing.assert_array_almost_equal(
                bt[0], np.array([score_hi_expected] * 2)
            )
            assert isinstance(bt[1], np.ndarray)
            np.testing.assert_array_almost_equal(
                bt[1], np.array([score_lo_expected] * 2)
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [
                ([metrics.mape], [0.0, 100.0]),
                ([metrics.mape, metrics.mape], [0.0, 100.0]),
                ([metrics.f1], [1.0, 0.0]),
                ([metrics.f1, metrics.precision], [1.0, 0.0]),
            ],
        ),
    )
    def test_output_multi_series_hfc_lpo_false_different_n_fcs(self, config):
        """Tests backtest based on historical forecasts generated on multiple `series` with
        last_points_only=False, and the historical forecasts have different lengths
        """
        is_univariate, (metric, score_expected) = config
        score_hi_expected, score_lo_expected = score_expected
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [[y], [hfc, hfc]]
        y = [y, y]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        assert str(err.value).startswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=None, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        assert isinstance(bt[0], np.ndarray)
        assert isinstance(bt[1], np.ndarray)
        if not is_multi_metric:
            # inner shape expected: (n hist forecasts = 1,)
            np.testing.assert_array_almost_equal(bt[0], np.array([score_hi_expected]))
            # inner shape expected: (n hist forecasts = 2,)
            np.testing.assert_array_almost_equal(
                bt[1], np.array([score_lo_expected] * 2)
            )
        else:
            # inner shape expected: (n metrics = 2, n hist forecasts = 1)
            np.testing.assert_array_almost_equal(
                bt[0], np.array([[score_hi_expected] * 2])
            )
            # inner shape expected: (n metrics = 2, n hist forecasts = 2)
            np.testing.assert_array_almost_equal(
                bt[1], np.array([[score_lo_expected] * 2, [score_lo_expected] * 2])
            )

        # with reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=np.mean, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # inner type expected: 1 float
            assert bt == score_expected
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt[0], np.ndarray)
            np.testing.assert_array_almost_equal(
                bt[0], np.array([score_hi_expected] * 2)
            )
            assert isinstance(bt[1], np.ndarray)

    def test_backtest_forecasting(self):
        linear_series = lt(length=50)
        linear_series_int = TimeSeries.from_values(linear_series.values())
        linear_series_multi = linear_series.stack(linear_series)

        # univariate model + univariate series
        score = NaiveDrift().backtest(
            linear_series,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score == 1.0

        # univariate model + univariate series + historical_forecasts precalculated
        forecasts = NaiveDrift().historical_forecasts(
            linear_series,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            last_points_only=False,
        )
        precalculated_forecasts_score = NaiveDrift().backtest(
            linear_series,
            historical_forecasts=forecasts,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score == precalculated_forecasts_score

        # very large train length should not affect the backtest
        score = NaiveDrift().backtest(
            linear_series,
            train_length=10000,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score == 1.0

        # using several metric function should not affect the backtest
        score = NaiveDrift().backtest(
            linear_series,
            train_length=10000,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=[metrics.r2_score, metrics.mape],
        )
        np.testing.assert_almost_equal(score, np.array([1.0, 0.0]))

        # window of size 2 is too small for naive drift
        with pytest.raises(ValueError):
            NaiveDrift().backtest(
                linear_series,
                train_length=2,
                start=pd.Timestamp("20000201"),
                forecast_horizon=3,
                metric=metrics.r2_score,
            )

        # test that it also works for time series that are not Datetime-indexed
        score = NaiveDrift().backtest(
            linear_series_int, start=0.7, forecast_horizon=3, metric=metrics.r2_score
        )
        assert score == 1.0

        with pytest.raises(ValueError):
            NaiveDrift().backtest(
                linear_series,
                start=pd.Timestamp("20000218"),
                forecast_horizon=3,
                overlap_end=False,
            )
        NaiveDrift().backtest(
            linear_series, start=pd.Timestamp("20000217"), forecast_horizon=3
        )
        NaiveDrift().backtest(
            linear_series,
            start=pd.Timestamp("20000218"),
            forecast_horizon=3,
            overlap_end=True,
        )

        # Using forecast_horizon default value
        NaiveDrift().backtest(linear_series, start=pd.Timestamp("20000216"))
        NaiveDrift().backtest(
            linear_series, start=pd.Timestamp("20000217"), overlap_end=True
        )

        # Using an int or float value for start
        NaiveDrift().backtest(linear_series, start=30)
        NaiveDrift().backtest(linear_series, start=0.7, overlap_end=True)

        # Set custom train window length
        NaiveDrift().backtest(linear_series, train_length=10, start=30)

        # Using invalid start and/or forecast_horizon values
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=0.7, forecast_horizon=-1)
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=-0.7, forecast_horizon=1)

        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=100)
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=1.2)
        with pytest.raises(TypeError):
            NaiveDrift().backtest(linear_series, start="wrong type")
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, train_length=0, start=0.5)
        with pytest.raises(TypeError):
            NaiveDrift().backtest(linear_series, train_length="wrong type", start=0.5)

        with pytest.raises(ValueError):
            NaiveDrift().backtest(
                linear_series, start=49, forecast_horizon=2, overlap_end=False
            )

        # univariate model + multivariate series
        with pytest.raises(ValueError):
            FFT().backtest(
                linear_series_multi, start=pd.Timestamp("20000201"), forecast_horizon=3
            )

        # multivariate model + univariate series
        if TORCH_AVAILABLE:
            tcn_model = TCNModel(
                input_chunk_length=12,
                output_chunk_length=1,
                batch_size=1,
                n_epochs=1,
                **tfm_kwargs,
            )
            # cannot perform historical forecasts with `retrain=False` and untrained model
            with pytest.raises(ValueError):
                _ = tcn_model.historical_forecasts(
                    linear_series,
                    start=pd.Timestamp("20000125"),
                    forecast_horizon=3,
                    verbose=False,
                    last_points_only=True,
                    retrain=False,
                )

            pred = tcn_model.historical_forecasts(
                linear_series,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
                last_points_only=True,
            )
            assert pred.width == 1
            assert pred.end_time() == linear_series.end_time()

            # multivariate model + multivariate series
            # historical forecasts doesn't overwrite model object -> we can use different input dimensions
            tcn_model.backtest(
                linear_series_multi,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
            )

            # univariate model
            tcn_model = TCNModel(
                input_chunk_length=12,
                output_chunk_length=1,
                batch_size=1,
                n_epochs=1,
                **tfm_kwargs,
            )
            tcn_model.fit(linear_series, verbose=False)
            # univariate fitted model + multivariate series
            with pytest.raises(ValueError):
                tcn_model.backtest(
                    linear_series_multi,
                    start=pd.Timestamp("20000125"),
                    forecast_horizon=3,
                    verbose=False,
                    retrain=False,
                )

            tcn_model = TCNModel(
                input_chunk_length=12,
                output_chunk_length=3,
                batch_size=1,
                n_epochs=1,
                **tfm_kwargs,
            )
            pred = tcn_model.historical_forecasts(
                linear_series_multi,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
                last_points_only=True,
            )
            assert pred.width == 2
            assert pred.end_time() == linear_series.end_time()

    def test_backtest_multiple_series(self):
        series = [AirPassengersDataset().load(), MonthlyMilkDataset().load()]
        model = NaiveSeasonal(K=1)

        error = model.backtest(
            series,
            train_length=30,
            forecast_horizon=2,
            stride=1,
            retrain=True,
            last_points_only=False,
            verbose=False,
        )

        expected = [11.63104, 6.09458]
        assert len(error) == 2
        assert round(abs(error[0] - expected[0]), 4) == 0
        assert round(abs(error[1] - expected[1]), 4) == 0

    def test_backtest_regression(self, caplog):
        np.random.seed(4)

        gaussian_series = gt(mean=2, length=50)
        sine_series = st(length=50)
        features = gaussian_series.stack(sine_series)
        features_multivariate = (
            (gaussian_series + sine_series).stack(gaussian_series).stack(sine_series)
        )
        target = sine_series

        features = features.with_columns_renamed(
            features.components, ["Value0", "Value1"]
        )

        features_multivariate = features_multivariate.with_columns_renamed(
            features_multivariate.components, ["Value0", "Value1", "Value2"]
        )

        # univariate feature test
        score = LinearRegressionModel(
            lags=None, lags_future_covariates=[0, -1]
        ).backtest(
            series=target,
            future_covariates=features,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
            last_points_only=True,
        )
        assert score > 0.9

        # univariate feature test + train length
        score = LinearRegressionModel(
            lags=None, lags_future_covariates=[0, -1]
        ).backtest(
            series=target,
            future_covariates=features,
            start=pd.Timestamp("20000201"),
            train_length=20,
            forecast_horizon=3,
            metric=metrics.r2_score,
            last_points_only=True,
        )
        assert score > 0.9

        # Using an int or float value for start
        score = RandomForestModel(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features,
            start=30,
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score > 0.9

        score = RandomForestModel(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features,
            start=0.5,
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score > 0.9

        # Using a too small start value
        warning_expected = (
            "`start` position `{0}` corresponding to time `{1}` is before the first "
            "predictable/trainable historical forecasting point for series at index: 0. Using the first historical "
            "forecasting point `2000-01-15 00:00:00` that lies a round-multiple of `stride=1` ahead of `start`. "
            "To hide these warnings, set `show_warnings=False`."
        )
        caplog.clear()
        with caplog.at_level(logging.WARNING):
            _ = RandomForestModel(lags=12).backtest(
                series=target, start=0, forecast_horizon=3
            )
            assert warning_expected.format(0, target.start_time()) in caplog.text
        caplog.clear()

        with caplog.at_level(logging.WARNING):
            _ = RandomForestModel(lags=12).backtest(
                series=target, start=0.01, forecast_horizon=3
            )
            assert warning_expected.format(0.01, target.start_time()) in caplog.text
        caplog.clear()

        # Using RandomForestModel's start default value
        score = RandomForestModel(lags=12, random_state=0).backtest(
            series=target, forecast_horizon=3, start=0.5, metric=metrics.r2_score
        )
        assert score > 0.95

        # multivariate feature test
        score = RandomForestModel(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score > 0.94

        # multivariate feature test with train window 35
        score_35 = RandomForestModel(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            train_length=35,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        logger.info(
            "Score for multivariate feature test with train window 35 is: ", score_35
        )
        assert score_35 > 0.92

        # multivariate feature test with train window 45
        score_45 = RandomForestModel(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            train_length=45,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        logger.info(
            "Score for multivariate feature test with train window 45 is: ", score_45
        )
        assert score_45 > 0.94
        assert score_45 > score_35

        # multivariate with stride
        score = RandomForestModel(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
            last_points_only=True,
            stride=3,
        )
        assert score > 0.9

    @pytest.mark.parametrize(
        "config",
        product(
            [
                "det",
                "sampled",
                "lkl_params",
            ],  # prediction method
            [False, True],  # is multivariate
            [True, False],  # auto-regression
        ),
    )
    def test_backtest_classification(self, config, caplog):
        """Tests backtest for classification models with different prediction methods and configurations."""
        method, is_multivariate, use_autoreg = config
        random_state = 42

        min_score = 1.0
        n_reps = 10
        ocl = 1 if use_autoreg else 3
        if method == "sampled":
            # sampled predictions need more training data to be accurate
            n_reps = 50
            if use_autoreg:
                # auto-regression would need even more data
                min_score = 0.9
        elif method == "lkl_params":
            # likelihood parameters prediction cannot perform autoregression
            ocl = 3

        labels = [4, 3, 2, 1, 0]
        series = np.array(labels * n_reps)
        times = generate_index("2000-01-01", length=len(series), freq="D")
        series = TimeSeries.from_times_and_values(times=times, values=series)
        if is_multivariate:
            series = series.stack(series + 10)
            labels += [14, 13, 12, 11, 10]

        model = SKLearnClassifierModel(
            lags=1,
            output_chunk_length=ocl,
            model=LogisticRegression(),
            random_state=random_state,
        ).fit(series)
        bt_kwargs = {
            "series": series,
            "start": pd.Timestamp("20000201"),
            "forecast_horizon": 3,
            "retrain": False,
            "random_state": random_state,
        }
        if method == "sampled":
            bt_kwargs["num_samples"] = 100
        elif method == "lkl_params":
            bt_kwargs["predict_likelihood_parameters"] = True

        # base clf metric reduces label-specific scores -> works out of the box
        for lpo in [True, False]:
            score = model.backtest(metric=metrics.f1, last_points_only=lpo, **bt_kwargs)
            assert score >= min_score

        score_expected = np.array([min_score] * len(labels))

        # clf metric without label reduction works for `lpo=True` as all labels are present
        score = model.backtest(
            metric=metrics.f1,
            metric_kwargs={"label_reduction": None},
            last_points_only=True,
            **bt_kwargs,
        )
        assert score.shape == score_expected.shape
        assert np.all(score >= score_expected)

        # with `lpo=False` it raises a warning that the results might be inconsistent (we get 3 scores instead of
        # 5 because metrics on multiple series / forecasts do not know about cross forecast labels)
        warning_expected = (
            "Parameter `labels` is not set for classification metric `f1()`"
        )
        caplog.clear()
        with caplog.at_level(logging.WARNING):
            score = model.backtest(
                metric=metrics.f1,
                metric_kwargs={"label_reduction": None},
                last_points_only=False,
                **bt_kwargs,
            )
            score_expected_ = np.array([min_score] * 3 * series.width)
            assert score.shape == score_expected_.shape
            assert np.all(score >= score_expected_)
            assert warning_expected in caplog.text
        caplog.clear()

        # passing the `labels` parameter gives the expected results and no warning
        caplog.clear()
        with caplog.at_level(logging.WARNING):
            score = model.backtest(
                metric=metrics.f1,
                metric_kwargs={"label_reduction": None, "labels": labels},
                last_points_only=False,
                **bt_kwargs,
            )
            assert score.shape == score_expected.shape
            assert np.all(score >= score_expected)
        caplog.clear()

        # no need to check below with sampled predictions using auto-regression, as output is more random
        if method == "sampled" and use_autoreg:
            return

        # selecting a subset of labels works as well
        # sampled predictions are not perfect, so we cannot expect 1.0 score
        if not is_multivariate:
            # univariate doesn't contain label `12`
            scores_expected = [[1.0], [1.0, np.nan]]
        else:
            scores_expected = [[1.0], [1.0, 1.0]]
        for labels_, score_expected in zip([[1.0], [1.0, 12.0]], scores_expected):
            score = model.backtest(
                metric=metrics.f1,
                metric_kwargs={"label_reduction": None, "labels": labels_},
                last_points_only=False,
                **bt_kwargs,
            )
            np.testing.assert_array_almost_equal(score, score_expected)

        # confusion matrix works as well
        score = model.backtest(
            metric=metrics.confusion_matrix,
            metric_kwargs={"labels": labels},
            reduction=np.nansum,
            last_points_only=False,
            **bt_kwargs,
        )
        # only diagonal elements are non-zero (true positives, perfect prediction
        non_zero_idx = np.argwhere(score > 0)
        assert np.all(non_zero_idx[:, 0] == non_zero_idx[:, 1])

    @pytest.mark.parametrize("model_cls", [Theta, ARIMA])
    def test_backtest_bad_covariates(self, model_cls):
        """Passing unsupported covariate should raise an exception"""
        series = lt(start_value=1, end_value=10, length=31, dtype="float32")
        model = model_cls()
        bt_kwargs = {"start": -1, "start_format": "position", "show_warnings": False}
        model.backtest(series=series, **bt_kwargs)

        with pytest.raises(ValueError) as msg:
            model.backtest(series=series, past_covariates=series, **bt_kwargs)
        assert str(msg.value).startswith(
            "Model cannot be fit/trained with `past_covariates`."
        )
        if not model.supports_future_covariates:
            with pytest.raises(ValueError) as msg:
                model.backtest(series=series, future_covariates=series, **bt_kwargs)
            assert str(msg.value).startswith(
                "Model cannot be fit/trained with `future_covariates`."
            )

    def test_gridsearch(self):
        np.random.seed(1)

        dummy_series = get_dummy_series(ts_length=50)
        dummy_series_int_index = TimeSeries.from_values(dummy_series.values())

        theta_params = {"theta": list(range(3, 10))}
        assert compare_best_against_random(Theta, theta_params, dummy_series)
        assert compare_best_against_random(Theta, theta_params, dummy_series_int_index)
        assert compare_best_against_random(Theta, theta_params, dummy_series, stride=2)

        fft_params = {"nr_freqs_to_keep": [10, 50, 100], "trend": [None, "poly", "exp"]}
        assert compare_best_against_random(FFT, fft_params, dummy_series)

        es_params = {"seasonal_periods": list(range(5, 10))}
        assert compare_best_against_random(
            ExponentialSmoothing, es_params, dummy_series
        )

    def test_gridsearch_metric_score(self):
        np.random.seed(1)

        model_class = Theta
        params = {"theta": list(range(3, 6))}
        dummy_series = get_dummy_series(ts_length=50)

        best_model, _, score = model_class.gridsearch(
            params,
            series=dummy_series,
            forecast_horizon=10,
            stride=1,
            start=dummy_series.time_index[-21],
        )
        recalculated_score = best_model.backtest(
            series=dummy_series,
            start=dummy_series.time_index[-21],
            forecast_horizon=10,
            stride=1,
        )

        assert score == recalculated_score, "The metric scores should match"

    def test_gridsearch_random_search(self):
        np.random.seed(1)

        dummy_series = get_dummy_series(ts_length=50)

        param_range = list(range(10, 20))
        params = {"lags": param_range}

        model = RandomForestModel(lags=1)
        result = model.gridsearch(
            params, dummy_series, forecast_horizon=1, n_random_samples=5
        )

        assert isinstance(result[0], RandomForestModel)
        assert isinstance(result[1]["lags"], int)
        assert isinstance(result[2], float)
        assert min(param_range) <= result[1]["lags"] <= max(param_range)

    def test_gridsearch_n_random_samples_bad_arguments(self):
        dummy_series = get_dummy_series(ts_length=50)

        params = {"lags": list(range(1, 11)), "past_covariates": list(range(1, 11))}

        with pytest.raises(ValueError):
            RandomForestModel.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=-5
            )
        with pytest.raises(ValueError):
            RandomForestModel.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=105
            )
        with pytest.raises(ValueError):
            RandomForestModel.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=-24.56
            )
        with pytest.raises(ValueError):
            RandomForestModel.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=1.5
            )

    def test_gridsearch_n_random_samples(self):
        np.random.seed(1)

        params = {"lags": list(range(1, 11)), "past_covariates": list(range(1, 11))}

        params_cross_product = list(product(*params.values()))

        # Test absolute sample
        absolute_sampled_result = RandomForestModel._sample_params(
            params_cross_product, 10
        )
        assert len(absolute_sampled_result) == 10

        # Test percentage sample
        percentage_sampled_result = RandomForestModel._sample_params(
            params_cross_product, 0.37
        )
        assert len(percentage_sampled_result) == 37

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_gridsearch_n_jobs(self):
        """
        Testing that running gridsearch with multiple workers returns the same
        best_parameters as the single worker run.
        """

        rng_seed = 1

        np.random.seed(rng_seed)

        dummy_series = get_dummy_series(
            ts_length=100, lt_end_value=1, st_value_offset=0
        ).astype(np.float32)
        ts_train, ts_val = dummy_series.split_before(split_point=0.8)

        test_cases = [
            {
                "model": ARIMA,  # ExtendedForecastingModel
                "parameters": {"p": [18, 4], "q": [2, 3], "random_state": [rng_seed]},
            },
            {
                "model": BlockRNNModel,  # TorchForecastingModel
                "parameters": {
                    "input_chunk_length": [5, 10],
                    "output_chunk_length": [1, 3],
                    "n_epochs": [1, 5],
                    "random_state": [rng_seed],
                    "pl_trainer_kwargs": [tfm_kwargs["pl_trainer_kwargs"]],
                },
            },
        ]

        for test in test_cases:
            model = test["model"]
            parameters = test["parameters"]

            np.random.seed(rng_seed)
            _, best_params1, _ = model.gridsearch(
                parameters=parameters, series=ts_train, val_series=ts_val, n_jobs=1
            )

            np.random.seed(rng_seed)
            _, best_params2, _ = model.gridsearch(
                parameters=parameters, series=ts_train, val_series=ts_val, n_jobs=2
            )

            assert best_params1 == best_params2

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_gridsearch_multi(self):
        dummy_series = st(length=40, value_y_offset=10).stack(
            lt(length=40, end_value=20)
        )
        tcn_params = {
            "input_chunk_length": [12],
            "output_chunk_length": [3],
            "n_epochs": [1],
            "batch_size": [1],
            "kernel_size": [2, 3, 4],
            "pl_trainer_kwargs": [tfm_kwargs["pl_trainer_kwargs"]],
        }
        TCNModel.gridsearch(
            tcn_params, dummy_series, forecast_horizon=3, metric=metrics.mape
        )

    @pytest.mark.parametrize(
        "model_cls,parameters",
        zip([NaiveSeasonal, ARIMA], [{"K": [1, 2]}, {"p": [18, 4]}]),
    )
    def test_gridsearch_bad_covariates(self, model_cls, parameters):
        """Passing unsupported covariate should raise an exception"""
        dummy_series = get_dummy_series(
            ts_length=100, lt_end_value=1, st_value_offset=0
        ).astype(np.float32)

        ts_train, ts_val = dummy_series.split_before(split_point=0.8)

        bt_kwargs = {"start": -1, "start_format": "position", "show_warnings": False}

        model = model_cls()
        model_cls.gridsearch(
            parameters=parameters, series=ts_train, val_series=ts_val, **bt_kwargs
        )

        with pytest.raises(ValueError) as msg:
            model_cls.gridsearch(
                parameters=parameters,
                series=ts_train,
                past_covariates=dummy_series,
                val_series=ts_val,
                **bt_kwargs,
            )
        assert str(msg.value).startswith(
            "Model cannot be fit/trained with `past_covariates`."
        )
        if not model.supports_future_covariates:
            with pytest.raises(ValueError) as msg:
                model_cls.gridsearch(
                    parameters=parameters,
                    series=ts_train,
                    future_covariates=dummy_series,
                    val_series=ts_val,
                    **bt_kwargs,
                )
            assert str(msg.value).startswith(
                "Model cannot be fit/trained with `future_covariates`."
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product([True, False], [True, False]),
    )
    def test_gridsearch_sample_weight(self, config):
        """check that passing sample weights work and that it yields different results than without sample weights."""
        manual_weight, use_val_series = config
        ts = AirPassengersDataset().load()
        if manual_weight:
            sample_weight = np.linspace(0, 1, len(ts))
            sample_weight = ts.with_values(np.expand_dims(sample_weight, -1))
        else:
            sample_weight = "linear"

        parameters = {"lags": [3], "output_chunk_length": [1]}
        start_kwargs = {"start": -1, "start_format": "position"}
        gs_kwargs = {"val_series": ts} if use_val_series else {"forecast_horizon": 1}
        gs_non_weighted = LinearRegressionModel.gridsearch(
            parameters, series=ts[:-1], **start_kwargs, **gs_kwargs
        )[-1]

        gs_weighted = LinearRegressionModel.gridsearch(
            parameters,
            series=ts[:-1],
            sample_weight=sample_weight,
            **start_kwargs,
            **gs_kwargs,
        )[-1]

        # check that the predictions are different
        assert gs_weighted != gs_non_weighted

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [
                metrics.ase,
                metrics.mase,
            ],
            [1, 2],
        ),
    )
    def test_scaled_metrics(self, config):
        """Tests backtest for scaled metrics based on historical forecasts generated on a sequence
        `series` with last_points_only=False"""
        metric, m = config
        y = lt(length=20)
        hfc = lt(length=10, start=y.start_time() + 10 * y.freq)
        y = [y, y]
        hfc = [[hfc, hfc], [hfc]]

        model = NaiveDrift()
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=False,
            reduction=None,
            metric_kwargs={"m": m},
        )
        assert isinstance(bts, list) and len(bts) == 2

        bt_expected = metric(y[0], hfc[0][0], insample=y[0], m=m)
        for bt_list in bts:
            for bt in bt_list:
                np.testing.assert_array_almost_equal(bt, bt_expected)

    @pytest.mark.parametrize(
        "metric",
        [
            [metrics.mae],  # mae does not support time_reduction
            [metrics.mae, metrics.ae],  # ae supports time_reduction
        ],
    )
    def test_metric_kwargs(self, metric):
        """Tests backtest with different metric_kwargs based on historical forecasts generated on a sequence
        `series` with last_points_only=False"""
        y = lt(length=20)
        y = y.stack(y + 1.0)
        hfc = lt(length=10, start=y.start_time() + 10 * y.freq)
        hfc = hfc.stack(hfc + 1.0)
        y = [y, y]
        hfc = [[hfc, hfc], [hfc]]

        metric_kwargs = [{"component_reduction": np.median}]
        if len(metric) > 1:
            # give metric specific kwargs
            metric_kwargs.append({
                "component_reduction": np.median,
                "time_reduction": np.mean,
            })

        model = NaiveDrift()
        # backtest should fail with invalid metric kwargs (mae does not support time reduction)
        with pytest.raises(TypeError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                metric=metric,
                last_points_only=False,
                reduction=None,
                metric_kwargs={
                    "component_reduction": np.median,
                    "time_reduction": np.mean,
                },
            )
        assert "unexpected keyword argument 'time_reduction'" in str(err.value)

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=False,
            reduction=None,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2

        # `ae` with time and component reduction is equal to `mae` with component reduction
        bt_expected = metrics.mae(y[0], hfc[0][0], component_reduction=np.median)
        for bt_list in bts:
            for bt in bt_list:
                np.testing.assert_array_almost_equal(bt, bt_expected)

        def time_reduced_metric(*args, **kwargs):
            return metrics.ae(*args, **kwargs, time_reduction=np.mean)

        # check that single kwargs can be used for all metrics if params are supported
        metric = [metric[0], time_reduced_metric]
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=False,
            reduction=None,
            metric_kwargs=metric_kwargs[0],
        )
        assert isinstance(bts, list) and len(bts) == 2

        # `ae` with time and component reduction is equal to `mae` with component reduction
        bt_expected = metrics.mae(y[0], hfc[0][0], component_reduction=np.median)
        for bt_list in bts:
            for bt in bt_list:
                np.testing.assert_array_almost_equal(bt, bt_expected)

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [
                [metrics.mae],  # mae does not support time_reduction
                [metrics.mae, metrics.ae],  # ae supports time_reduction
                [metrics.miw],  # quantile interval metric
                [metrics.miw, metrics.iw],
            ],
            [True, False],  # last_points_only
        ),
    )
    def test_metric_quantiles_lpo(self, config):
        """Tests backtest with quantile and quantile interval metrics from expected probabilistic or quantile
        historical forecasts."""
        metric, lpo = config
        is_interval_metric = metric[0].__name__ == "miw"

        q = [0.05, 0.5, 0.60, 0.95]
        q_interval = [(0.05, 0.50), (0.50, 0.60), (0.60, 0.95), (0.05, 0.60)]

        y = lt(length=20)
        y = y.stack(y + 1.0)
        hfc = TimeSeries.from_times_and_values(
            times=generate_index(start=y.start_time() + 10 * y.freq, length=10),
            values=np.random.random((10, 1, 100)),
        )
        hfc = hfc.stack(hfc + 1.0)
        y = [y, y]
        if lpo:
            hfc = [hfc, hfc]
        else:
            hfc = [[hfc, hfc], [hfc]]

        metric_kwargs = [{"component_reduction": np.median}]
        if not is_interval_metric:
            metric_kwargs[0]["q"] = q
        else:
            metric_kwargs[0]["q_interval"] = q_interval
        if len(metric) > 1:
            # give metric specific kwargs
            metric_kwargs2 = {
                "component_reduction": np.median,
                "time_reduction": np.mean,
            }
            if not is_interval_metric:
                metric_kwargs2["q"] = q
            else:
                metric_kwargs2["q_interval"] = q_interval
            metric_kwargs.append(metric_kwargs2)

        model = NaiveDrift()

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=None,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        if lpo:
            bts = [[bt] for bt in bts]
        # `ae` with time and component reduction is equal to `mae` with component reduction
        hfc_single = hfc[0][0] if not lpo else hfc[0]
        q_kwargs = {"q": q} if not is_interval_metric else {"q_interval": q_interval}
        bt_expected = metric[0](
            y[0], hfc_single, component_reduction=np.median, **q_kwargs
        )
        shape_expected = (len(q),)
        if len(metric) > 1:
            bt_expected = np.concatenate([bt_expected[:, None]] * 2, axis=1)
            shape_expected += (len(metric),)
        for bt_list in bts:
            for bt in bt_list:
                assert bt.shape == shape_expected
                np.testing.assert_array_almost_equal(bt, bt_expected)

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=np.mean,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        for bt in bts:
            assert bt.shape == shape_expected
            np.testing.assert_array_almost_equal(bt, bt_expected)

        def time_reduced_metric(*args, **kwargs):
            metric_f = metrics.iw if is_interval_metric else metrics.ae
            return metric_f(*args, **kwargs, time_reduction=np.mean)

        # check that single kwargs can be used for all metrics if params are supported
        metric = [metric[0], time_reduced_metric]
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=None,
            metric_kwargs=metric_kwargs[0],
        )
        assert isinstance(bts, list) and len(bts) == 2
        if lpo:
            bts = [[bt] for bt in bts]
        # `ae` / `miw` with time and component reduction is equal to `mae` / `miw` with component reduction
        bt_expected = metric[0](
            y[0], hfc_single, component_reduction=np.median, **q_kwargs
        )
        bt_expected = np.concatenate([bt_expected[:, None]] * 2, axis=1)
        shape_expected = (len(q), len(metric))
        for bt_list in bts:
            for bt in bt_list:
                assert bt.shape == shape_expected
                np.testing.assert_array_almost_equal(bt, bt_expected)

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=np.mean,
            metric_kwargs=metric_kwargs[0],
        )
        assert isinstance(bts, list) and len(bts) == 2
        for bt in bts:
            assert bt.shape == shape_expected
            np.testing.assert_array_almost_equal(bt, bt_expected)

        # without component reduction
        metric_kwargs = {"component_reduction": None}
        if not is_interval_metric:
            metric_kwargs["q"] = q
        else:
            metric_kwargs["q_interval"] = q_interval
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=None,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        if lpo:
            bts = [[bt] for bt in bts]

        # `ae` / `iw` with time and no component reduction is equal to `mae` / `miw` without component reduction
        bt_expected = metric[0](y[0], hfc_single, **metric_kwargs)
        bt_expected = np.concatenate([bt_expected[:, None]] * 2, axis=1)
        shape_expected = (len(q) * y[0].width, len(metric))
        for bt_list in bts:
            for bt in bt_list:
                assert bt.shape == shape_expected
                np.testing.assert_array_almost_equal(bt, bt_expected)

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=np.mean,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        for bt in bts:
            assert bt.shape == shape_expected
            np.testing.assert_array_almost_equal(bt, bt_expected)

    @pytest.mark.parametrize(
        "config",
        product([True, False], [True, False]),
    )
    def test_backtest_sample_weight(self, config):
        """check that passing sample weights work and that it yields different results than without sample weights."""
        manual_weight, multi_series = config
        ts = AirPassengersDataset().load()
        if manual_weight:
            sample_weight = np.linspace(0, 1, len(ts))
            sample_weight = ts.with_values(np.expand_dims(sample_weight, -1))
        else:
            sample_weight = "linear"

        if multi_series:
            ts = [ts] * 2
            sample_weight = [sample_weight] * 2 if manual_weight else sample_weight

        model = LinearRegressionModel(lags=3, output_chunk_length=1)
        start_kwargs = {"start": -1, "start_format": "position"}
        bt_non_weighted = model.backtest(series=ts, **start_kwargs)

        model = LinearRegressionModel(lags=3, output_chunk_length=1)
        bt_weighted = model.backtest(
            series=ts, sample_weight=sample_weight, **start_kwargs
        )

        if not multi_series:
            bt_weighted = [bt_weighted]
            bt_non_weighted = [bt_non_weighted]

        # check that the predictions are different
        for bt_nw, bt_w in zip(bt_non_weighted, bt_weighted):
            assert bt_w != bt_nw



import logging
from itertools import product
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from sklearn.base import BaseEstimator
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.metrics import f1_score
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import Tags

from darts.logging import get_logger
from darts.models import (
    CatBoostClassifierModel,
    LightGBMClassifierModel,
    SKLearnClassifierModel,
    XGBClassifierModel,
)
from darts.models.forecasting.sklearn_model import SKLearnModelWithCategoricalFeatures
from darts.tests.conftest import CB_AVAILABLE, LGBM_AVAILABLE, XGB_AVAILABLE
from darts.timeseries import TimeSeries
from darts.utils import timeseries_generation as tg
from darts.utils.likelihood_models.base import LikelihoodType
from darts.utils.likelihood_models.sklearn import (
    ClassProbabilityLikelihood,
    _get_likelihood,
)
from darts.utils.multioutput import (
    MultiOutputClassifier,
    MultiOutputRegressor,
    get_multioutput_estimator_cls,
)
from darts.utils.utils import ModelType

logger = get_logger(__name__)


def process_model_list(classifiers):
    for clf, kwargs in classifiers:
        if issubclass(clf, BaseEstimator):
            yield (SKLearnClassifierModel, {"model": clf(**kwargs), "random_state": 42})
        else:
            yield (clf, kwargs)


def train_test_split(series, split_ts):
    """
    Splits all provided TimeSeries instances into train and test sets according to the provided timestamp.

    Parameters
    ----------
    features : TimeSeries
        Feature TimeSeries instances to be split.
    target : TimeSeries
        Target TimeSeries instance to be split.
    split_ts : TimeStamp
        Time stamp indicating split point.

    Returns
    -------
    TYPE
        4-tuple of the form (train_features, train_target, test_features, test_target)
    """
    if isinstance(series, TimeSeries):
        return series.split_after(split_ts)
    else:
        return list(zip(*[ts.split_after(split_ts) for ts in series]))


class TestClassifierModel:
    np.random.seed(42)

    # shift sines to positive values so that they can be used as target for classification with classes [0, 1, 2]
    sine_univariate1 = tg.sine_timeseries(length=100) + 1
    sine_univariate2 = tg.sine_timeseries(length=100, value_phase=1.5705) + 1
    sine_univariate3 = tg.sine_timeseries(length=100, value_phase=0.78525) + 1
    sine_univariate4 = tg.sine_timeseries(length=100, value_phase=0.392625) + 1
    sine_univariate5 = tg.sine_timeseries(length=100, value_phase=0.1963125) + 1
    sine_univariate6 = tg.sine_timeseries(length=100, value_phase=0.09815625) + 1

    sine_univariate1_cat = sine_univariate1.map(lambda x: np.round(x))
    sine_univariate2_cat = sine_univariate2.map(lambda x: np.round(x))
    sine_univariate3_cat = sine_univariate3.map(lambda x: np.round(x))

    sine_multivariate1_cat = sine_univariate1_cat.stack(sine_univariate2_cat)
    sine_multivariate2_cat = sine_univariate2_cat.stack(sine_univariate3_cat)

    sine_multivariate1 = sine_univariate2.stack(sine_univariate3)
    sine_multivariate2 = sine_univariate1.stack(sine_univariate3)

    sine_multiseries1_cat = [
        sine_univariate1_cat,
        sine_univariate2_cat,
        sine_univariate3_cat,
    ]
    sine_multiseries1 = [sine_univariate4, sine_univariate5, sine_univariate6]

    sine_multivariate_multiseries = [sine_multivariate1, sine_multivariate2]
    sine_multivariate_multiseries_cat = [sine_multivariate1_cat, sine_multivariate2_cat]

    classifiers = [
        (LogisticRegression, {}),
        (KNeighborsClassifier, {"n_neighbors": 3}),
        (SVC, {"gamma": 2, "C": 1, "random_state": 42, "probability": True}),
        (GaussianProcessClassifier, {"kernel": 1.0 * RBF(1.0), "random_state": 42}),
        (DecisionTreeClassifier, {"max_depth": 5, "random_state": 42}),
        (
            RandomForestClassifier,
            {"max_depth": 5, "n_estimators": 10, "max_features": 1, "random_state": 42},
        ),
        (MLPClassifier, {"alpha": 1, "max_iter": 1000, "random_state": 42}),
        (AdaBoostClassifier, {"random_state": 42}),
        (GaussianNB, {}),
    ]

    models_accuracies = [
        1,  # SKLearnClassifierModel
        1,  # KNeighborsClassifier
        1,  # SVC
        1,  # GaussianProcessClassifier
        1,  # DecisionTreeClassifier
        1,  # RandomForestClassifier
        1,  # MLPClassifier
        1,  # AdaBoostClassifier
        1,  # GaussianNB
    ]

    models_multioutput = [
        False,  # SKLearnClassifierModel
        True,  # KNeighborsClassifier
        False,  # SVC
        False,  # GaussianProcessClassifier
        True,  # DecisionTreeClassifier
        True,  # RandomForestClassifier
        False,  # MLPClassifier
        False,  # AdaBoostClassifier
        False,  # GaussianNB
    ]

    if XGB_AVAILABLE:
        classifiers.append((
            XGBClassifierModel,
            {
                "n_estimators": 1,
                "max_depth": 1,
                "max_leaves": 1,
                "random_state": 42,
            },
        ))
        models_accuracies.append(1)
        models_multioutput.append(False)

    if LGBM_AVAILABLE:
        classifiers.append((
            LightGBMClassifierModel,
            {
                "n_estimators": 1,
                "max_depth": 1,
                "num_leaves": 2,
                "verbosity": -1,
                "random_state": 42,
            },
        ))
        models_accuracies.append(1)
        models_multioutput.append(False)

    if CB_AVAILABLE:
        classifiers.append((
            CatBoostClassifierModel,
            {
                "iterations": 1,
                "depth": 1,
                "verbose": -1,
                "random_state": 42,
            },
        ))
        models_accuracies.append(1)
        models_multioutput.append(False)

    @pytest.mark.parametrize("clf_params", process_model_list(classifiers))
    def test_init_classifier(self, clf_params):
        clf, kwargs = clf_params
        model = clf(lags_past_covariates=5, **kwargs)
        assert model.model is not None
        assert isinstance(model.likelihood, ClassProbabilityLikelihood)

        # accepts only classifier
        with pytest.raises(ValueError) as err:
            SKLearnClassifierModel(model=LinearRegression(), lags=1)
        assert (
            str(err.value)
            == "`SKLearnClassifierModel` must be initialized with a classifier `model`."
        )

    @pytest.mark.parametrize("clf_params", process_model_list(classifiers))
    def test_univariate_class_labels(self, clf_params):
        clf, kwargs = clf_params
        model = clf(lags_past_covariates=5, **kwargs)

        # accessing classes_ before training return None
        assert model.class_labels is None

        # training the model
        model.fit(
            series=self.sine_univariate1_cat, past_covariates=self.sine_univariate1
        )
        # `class_labels` is a list of component-specific numpy array (univariate = length 1)
        assert isinstance(model.class_labels, list)
        assert len(model.class_labels) == 1
        assert isinstance(model.class_labels[0], np.ndarray)
        assert set(np.unique(self.sine_univariate1_cat.values())) == {0, 1, 2}
        assert (model.class_labels[0] == [0, 1, 2]).all()

    @pytest.mark.parametrize("clf_params", process_model_list(classifiers))
    def test_multiclass_class_labels(self, clf_params):
        clf, kwargs = clf_params
        model = clf(lags_past_covariates=5, **kwargs)

        if XGB_AVAILABLE and issubclass(clf, XGBClassifierModel):
            # XGB requires class labels to be consecutive from 0
            multivariate_cat_diff_labels = self.sine_univariate1_cat.stack(
                self.sine_univariate1_cat
            )
            expected_classes = [np.array([0, 1, 2]), np.array([0, 1, 2])]
            classes = {0, 1, 2}
        else:
            multivariate_cat_diff_labels = self.sine_univariate1_cat.stack(
                self.sine_univariate1_cat + 3
            )
            expected_classes = [np.array([0, 1, 2]), np.array([3, 4, 5])]
            classes = {0, 1, 2, 3, 4, 5}

        assert set(np.unique(multivariate_cat_diff_labels.values())) == classes
        model.fit(
            series=multivariate_cat_diff_labels, past_covariates=self.sine_univariate1
        )
        # `class_labels` is a list of component-specific numpy array (multivariate = length 2)
        # for both MultiOutputClassifier and native multi-class classifiers
        assert isinstance(model.class_labels, list)
        assert len(model.class_labels) == len(expected_classes)
        assert isinstance(model.class_labels[0], np.ndarray)
        # check that classes correspond to the classes in the series
        assert np.array([
            (c1 == c2).all() for c1, c2 in zip(model.class_labels, expected_classes)
        ]).all()

    @pytest.mark.parametrize(
        "clf_params",
        [(SKLearnClassifierModel, {})]
        + (
            [
                (
                    XGBClassifierModel,
                    {
                        "n_estimators": 1,
                        "max_depth": 1,
                        "max_leaves": 1,
                        "random_state": 42,
                    },
                ),
            ]
            if XGB_AVAILABLE
            else []
        ),
    )
    def test_error_on_different_classes_for_same_component(self, clf_params):
        """check that estimators for the same component see the same labels"""
        clf, kwargs = clf_params
        # only one estimator see the last labels of the series thus the estimator due to output_chunk_length=2
        # for the same component won't see the same labels
        series = TimeSeries.from_values(np.array([0, 0, 0, 1, 2]), columns=["comp1"])
        model = clf(lags=1, output_chunk_length=2, **kwargs)

        with pytest.raises(ValueError) as err:
            model.fit(series=series)
        assert str(err.value).startswith(
            "Estimators for the same target component received different class labels "
            "during training. In most cases this occurs for shorter target series. "
        )

        # if only one estimator due to multi_models=False, then there are no issue
        single_model = clf(lags=1, output_chunk_length=2, multi_models=False, **kwargs)
        single_model.fit(series=series)

        # if same labels per component it does not raise an error
        series = TimeSeries.from_values(np.array([0, 0, 0, 1, 1]), columns=["comp1"])
        model.fit(series=series)

    @pytest.mark.parametrize("clf_params", process_model_list(classifiers))
    def test_optional_static_covariates(self, clf_params):
        """adding static covariates to lagged data logic is tested in
        `darts.tests.utils.data.tabularization.test_add_static_covariates`
        """
        series = self.sine_univariate1_cat.with_static_covariates(
            pd.DataFrame({"a": [1]})
        ).astype(np.int32)

        # training model with static covs and predicting without will raise an error
        clf, kwargs = clf_params
        # with `use_static_covariates=False`, static covariates are ignored and prediction works
        model = clf(lags=4, use_static_covariates=False, **kwargs)
        model.fit(series.with_static_covariates(None))
        assert not model.uses_static_covariates
        assert model._static_covariates_shape is None
        preds = model.predict(n=2, series=series)
        np.testing.assert_almost_equal(
            preds.static_covariates.values,
            series.static_covariates.values,
        )

        # with `use_static_covariates=True`, static covariates are included
        model = clf(lags=4, use_static_covariates=True, **kwargs)
        model.fit([series, series])
        assert model.uses_static_covariates
        assert model._static_covariates_shape == series.static_covariates.shape
        preds = model.predict(n=2, series=[series, series])
        for pred in preds:
            pred.static_covariates.equals(series.static_covariates)

    def helper_test_models_accuracy(
        self,
        series,
        past_covariates,
        min_f1_model,
        model_params,
        idx,
        multi_models,
        output_chunk_length,
    ):
        multi_series = not isinstance(series, TimeSeries)
        # for every model, test whether it predicts the target with a minimum f1 score
        train_series, test_series = train_test_split(series, 70)
        train_past_covariates, _ = train_test_split(past_covariates, 70)
        model, kwargs = model_params
        model_instance = model(
            lags=12,
            lags_past_covariates=2,
            output_chunk_length=output_chunk_length,
            multi_models=multi_models,
            **kwargs,
        )
        model_instance.fit(series=train_series, past_covariates=train_past_covariates)
        prediction = model_instance.predict(
            n=len(test_series) if not multi_series else len(test_series[0]),
            series=train_series,
            past_covariates=past_covariates,
        )

        # flatten in case of multivariate prediction
        if multi_series:
            prediction = np.array([p.values().flatten() for p in prediction]).flatten()
            test_series = np.array([
                ts.values().flatten() for ts in test_series
            ]).flatten()
        else:
            prediction = prediction.values().flatten()
            test_series = test_series.values().flatten()

        current_f1 = f1_score(test_series, prediction, average=None).mean()
        mean_f1 = np.mean(current_f1)
        assert mean_f1 <= min_f1_model[idx], (
            f"{str(model_instance)} model was not able to predict data as well as expected. "
            f"A mean f1 score of {mean_f1} was recorded."
        )

    @pytest.mark.parametrize(
        "config",
        product(
            zip(
                process_model_list(classifiers),
                range(len(list(process_model_list(classifiers)))),
            ),
            [True, False],
            [1, 5],
        ),
    )
    def test_models_accuracy_univariate(self, config):
        (model, idx), multi_models, output_chunk_length = config
        # for every model, and different output_chunk_lengths test whether it predicts the univariate time series
        # as well as expected, accuracies are defined at the top of the class
        self.helper_test_models_accuracy(
            self.sine_univariate1_cat,
            self.sine_univariate2,
            self.models_accuracies,
            model,
            idx,
            multi_models,
            output_chunk_length,
        )

    @pytest.mark.parametrize(
        "config",
        product(
            zip(
                process_model_list(classifiers),
                range(len(list(process_model_list(classifiers)))),
            ),
            [True, False],
            [1, 5],
        ),
    )
    def test_models_accuracy_multivariate(self, config):
        (model, idx), multi_models, output_chunk_length = config
        # for every model, and different output_chunk_lengths test whether it predicts the multivariate time series
        # as well as expected, accuracies are defined at the top of the class
        self.helper_test_models_accuracy(
            self.sine_multivariate1_cat,
            self.sine_multivariate1,
            self.models_accuracies,
            model,
            idx,
            multi_models,
            output_chunk_length,
        )

    @pytest.mark.parametrize(
        "config",
        product(
            zip(
                process_model_list(classifiers),
                range(len(list(process_model_list(classifiers)))),
            ),
            [True, False],
            [1, 5],
        ),
    )
    def test_models_accuracy_multiseries_multivariate(self, config):
        (model, idx), multi_models, ocl = config
        # for every model, and different output_chunk_lengths test whether it predicts the multiseries, multivariate
        # time series as well as expected, accuracies are defined at the top of the class
        self.helper_test_models_accuracy(
            self.sine_multiseries1_cat,
            self.sine_multiseries1,
            self.models_accuracies,
            model,
            idx,
            multi_models,
            ocl,
        )

    @pytest.mark.parametrize(
        "model_params",
        zip(process_model_list(classifiers), models_multioutput),
    )
    def test_multioutput_wrapper(self, model_params):
        """Check that with output_chunk_length=1, wrapping in MultiOutputClassifier occurs only when necessary"""
        (model_cls, kwargs), supports_multioutput_natively = model_params
        model = model_cls(
            lags_past_covariates=1,
            **kwargs,
        )

        def check_only_non_native_are_wrapped(model, supports_multioutput_natively):
            if supports_multioutput_natively:
                assert not isinstance(model.model, MultiOutputClassifier)
                # single estimator is responsible for both components
                assert (
                    model.model
                    == model.get_estimator(horizon=0, target_dim=0)
                    == model.get_estimator(horizon=0, target_dim=1)
                )
            else:
                assert isinstance(model.model, MultiOutputClassifier)
                # one estimator (sub-model) per component
                assert model.get_estimator(
                    horizon=0, target_dim=0
                ) != model.get_estimator(horizon=0, target_dim=1)

        # univariate should not be wrapped in MultiOutputRegressor
        model.fit(
            series=self.sine_univariate1_cat, past_covariates=self.sine_multivariate1
        )
        assert not isinstance(model.model, MultiOutputClassifier)

        model = model.untrained_model()
        # multivariate should be wrapped in MultiOutputRegressor only if not natively supported
        model.fit(
            series=self.sine_multivariate1_cat, past_covariates=self.sine_multivariate1
        )
        check_only_non_native_are_wrapped(model, supports_multioutput_natively)

        model = model.untrained_model()
        # multi-series with same component should not be wrapped in MultiOutputRegressor
        model.fit(
            series=self.sine_multiseries1_cat, past_covariates=self.sine_multiseries1
        )
        assert not isinstance(model.model, MultiOutputClassifier)

        model = model.untrained_model()
        # multi-series with multivariate should be wrapped in MultiOutputRegressor only if not natively supported
        model.fit(
            series=self.sine_multivariate_multiseries_cat,
            past_covariates=self.sine_multivariate_multiseries,
        )
        check_only_non_native_are_wrapped(model, supports_multioutput_natively)

    @pytest.mark.parametrize(
        "config",
        product(
            [
                model_config
                for multi_out, model_config in zip(
                    models_multioutput, process_model_list(classifiers)
                )
                if not multi_out
            ],
            [1, 2],
            [True, False],
        ),
    )
    def test_multioutput_validation(self, config):
        """Check that models not supporting multi-output are properly wrapped when ocl>1"""
        (model_cls, model_kwargs), ocl, multi_models = config
        train, val = self.sine_univariate1_cat.split_after(0.6)
        model = model_cls(
            **model_kwargs, lags=4, output_chunk_length=ocl, multi_models=multi_models
        )
        model.fit(series=train, val_series=val)
        if model.output_chunk_length > 1 and model.multi_models:
            assert isinstance(model.model, MultiOutputClassifier)
        else:
            assert not isinstance(model.model, MultiOutputClassifier)

    @pytest.mark.parametrize(
        "config", product(process_model_list(classifiers), [True, False])
    )
    def test_models_runnability(self, config):
        (model_cls, kwargs), multi_models = config

        train_y, test_y = self.sine_univariate1_cat.split_before(0.7)

        # testing past covariates
        model_instance = model_cls(
            lags_future_covariates=(0, 3),
            lags_past_covariates=None,
            multi_models=multi_models,
            **kwargs,
        )
        with pytest.raises(ValueError) as err:
            # testing lags_past_covariates None but past_covariates during training
            model_instance.fit(
                series=self.sine_univariate1_cat,
                past_covariates=self.sine_multivariate1,
                future_covariates=self.sine_multivariate1,
            )
        assert str(err.value).startswith(
            "`past_covariates` not None in `fit()` method call"
        )

        model_instance = model_cls(
            lags_past_covariates=3, multi_models=multi_models, **kwargs
        )
        with pytest.raises(ValueError) as err:
            # testing lags_past_covariates but no past_covariates during fit
            model_instance.fit(series=self.sine_univariate1_cat)
        assert str(err.value).startswith(
            "`past_covariates` is None in `fit()` method call"
        )

        # testing future_covariates
        model_instance = model_cls(
            lags_past_covariates=4,
            lags_future_covariates=None,
            multi_models=multi_models,
            **kwargs,
        )
        with pytest.raises(ValueError) as err:
            # testing lags_future_covariates None but future_covariates during training
            model_instance.fit(
                series=self.sine_univariate1_cat,
                past_covariates=self.sine_multivariate1,
                future_covariates=self.sine_multivariate1,
            )
        assert str(err.value).startswith(
            "`future_covariates` not None in `fit()` method call"
        )

        model_instance = model_cls(
            lags_past_covariates=4,
            lags_future_covariates=(0, 3),
            multi_models=multi_models,
            **kwargs,
        )
        with pytest.raises(ValueError) as err:
            # testing lags_future_covariates but no future_covariates during fit
            model_instance.fit(
                series=self.sine_univariate1_cat,
                past_covariates=self.sine_multivariate1,
            )
        assert str(err.value).startswith(
            "`future_covariates` is None in `fit()` method call"
        )

        # testing input_dim
        model_instance = model_cls(
            lags_past_covariates=2, multi_models=multi_models, **kwargs
        )
        model_instance.fit(
            series=train_y,
            past_covariates=self.sine_univariate1.stack(self.sine_univariate1)[
                : train_y.end_time()
            ],
        )

        assert model_instance.input_dim == {
            "target": 1,
            "past": 2,
            "future": None,
        }

        with pytest.raises(ValueError) as err:
            _ = model_instance.predict(n=model_instance.output_chunk_length + 1)
        assert str(err.value).startswith("The `past_covariates` are not long enough.")

        # while it should work with n = output_chunk_length
        prediction = model_instance.predict(n=model_instance.output_chunk_length)
        assert len(prediction) == 1

        # test wrong covariates dimensionality
        with pytest.raises(ValueError) as err:
            _ = prediction = model_instance.predict(
                n=model_instance.output_chunk_length,
                past_covariates=self.sine_univariate1,
            )
        assert str(err.value).startswith(
            "The number of components of the target series and the covariates"
        )

    @pytest.mark.parametrize("clf_params", process_model_list(classifiers))
    def test_labels_constraints(self, clf_params):
        clf, kwargs = clf_params
        model = clf(lags_past_covariates=2, **kwargs)

        # Classification forecasting models do not accept continuous labels
        # (different error messages depending on the classifier)
        with pytest.raises(ValueError):
            model.fit(
                series=self.sine_univariate1, past_covariates=self.sine_univariate1
            )

        # XGBClassifierModel require labels to be integers between 0 and n_classes
        if XGB_AVAILABLE and issubclass(clf, XGBClassifierModel):
            # negative labels
            with pytest.raises(ValueError) as err:
                model.fit(
                    series=self.sine_univariate1_cat - 5,
                    past_covariates=self.sine_univariate1,
                )
            assert str(err.value).endswith("Expected: [0 1 2], got [-5. -4. -3.]")

            # labels not between 0 and n_classes
            with pytest.raises(ValueError) as err:
                model.fit(
                    series=self.sine_univariate1_cat.map(
                        lambda x: np.where(x > 0, x + 1, x)
                    ),
                    past_covariates=self.sine_univariate1,
                )
            assert str(err.value).endswith("Expected: [0 1 2], got [0. 2. 3.]")

        # Single label
        if issubclass(clf, (SVC, GaussianProcessClassifier)) or isinstance(
            model.model, LogisticRegression
        ):
            # Model specific error message
            with pytest.raises(ValueError):
                model.fit(
                    series=self.sine_univariate1_cat - self.sine_univariate1_cat,
                    past_covariates=self.sine_univariate1,
                )

    @pytest.mark.parametrize("clf_params", process_model_list(classifiers))
    def test_warning_raised_on_lags_and_target_not_cat(self, clf_params, caplog):
        clf, kwargs = clf_params
        with caplog.at_level(logging.WARNING):
            model = clf(lags=2, **kwargs)
            if isinstance(model, SKLearnModelWithCategoricalFeatures):
                assert not caplog.messages
            else:
                assert any([
                    message.startswith(
                        "This model will treat target `series` lagged values as "
                        "numeric input features (and not categorical)."
                    )
                    for message in caplog.messages
                ])

    @pytest.mark.parametrize(
        "clf_params",
        [
            (model, config)
            for model, config in process_model_list(classifiers)
            if issubclass(model, SKLearnModelWithCategoricalFeatures)
        ],
    )
    def test_categorical_target_passed_to_fit_correctly(self, clf_params):
        clf, kwargs = clf_params
        model = clf(lags=2, lags_past_covariates=2, **kwargs)

        intercepted_args = {}
        original_fit = model.model.fit

        # catboost requires passing a verbose parameter;
        # Darts wrapper requires `verbose` to be in the fit signature for it to be passed to the underlying model
        if CB_AVAILABLE and isinstance(model, CatBoostClassifierModel):

            def intercept_fit_args(*args, verbose=False, **kwargs):
                intercepted_args["args"] = args
                intercepted_args["kwargs"] = kwargs
                return original_fit(*args, verbose=verbose, **kwargs)

        else:
            # other models do not require the verbose parameter
            def intercept_fit_args(*args, **kwargs):
                intercepted_args["args"] = args
                intercepted_args["kwargs"] = kwargs
                return original_fit(*args, **kwargs)

        # target is categorical by default for classifiers supporting it
        expected_cat_indices = [0, 1]
        with patch.object(
            model.model.__class__,
            "fit",
            side_effect=intercept_fit_args,
        ):
            model.fit(self.sine_univariate1_cat, self.sine_univariate2)

            cat_param_name = model._categorical_fit_param
            # check that the categorical index is passed to the fit method and that it is correct
            assert intercepted_args["kwargs"][cat_param_name] == expected_cat_indices

            if issubclass(clf, CatBoostClassifierModel):
                # check model has the correct categorical features
                assert model.model.get_cat_feature_indices() == expected_cat_indices

                X, y = intercepted_args["args"]
                # all categorical features should be encoded as integers
                assert isinstance(X, pd.DataFrame)
                for i, col in enumerate(X.columns):
                    assert X[col].dtype == (int if i in expected_cat_indices else float)
            elif issubclass(clf, LightGBMClassifierModel):
                X, y = intercepted_args["args"]
                assert isinstance(X, np.ndarray)
            else:
                assert False, f"{clf} need to be tested for fit arguments"

    def test_get_multioutput_estimator_cls(self):
        assert issubclass(
            get_multioutput_estimator_cls(ModelType.FORECASTING_CLASSIFIER),
            MultiOutputClassifier,
        )
        assert issubclass(
            get_multioutput_estimator_cls(ModelType.FORECASTING_REGRESSOR),
            MultiOutputRegressor,
        )

        with pytest.raises(ValueError) as err:
            get_multioutput_estimator_cls(model_type="invalid_type")
        assert (
            str(err.value)
            == "Model type must be one of `[ModelType.FORECASTING_REGRESSOR, ModelType.FORECASTING_CLASSIFIER]`. "
            "Received: `invalid_type`."
        )


@pytest.fixture(autouse=True)
def random():
    np.random.seed(0)


def generate_random_series_with_probabilities(
    multi_variate: bool, multi_series: bool, length=100
):
    """
    Generate categorical series with specific probabilities for each class
    """
    probas_component1 = np.array([0.1, 0.3, 0.6])
    labels_component1 = [0, 1, 2]

    probas_component2 = np.array([0.7, 0.3]) if multi_variate else None
    labels_component2 = [0, 1] if multi_variate else None

    columns = ["component1", "component2"] if multi_variate else ["component1"]

    def generate_samples():
        array = np.random.choice(
            labels_component1, size=length * 2, replace=True, p=probas_component1
        )

        if multi_variate:
            array = np.vstack((
                array,
                np.random.choice(
                    labels_component2,
                    size=length * 2,
                    replace=True,
                    p=probas_component2,
                ),
            )).T
        return array

    series = TimeSeries.from_values(generate_samples(), columns=columns)
    series_train, series_test = series.split_before(length)

    if multi_series:
        series2 = TimeSeries.from_values(generate_samples(), columns=columns)
        series2_train, series2_test = series2.split_before(length)
        series = [series, series2]
        series_train = [series_train, series2_train]
        series_test = [series_test, series2_test]

    return (
        series,
        series_train,
        series_test,
        probas_component1,
        labels_component1,
        probas_component2,
        labels_component2,
    )


class TestProbabilisticClassifierModels:
    probabilistic_classifiers = [
        (LogisticRegression, {}),
        (KNeighborsClassifier, {"n_neighbors": 10}),
        (GaussianProcessClassifier, {"kernel": 1.0 * RBF(1.0), "random_state": 42}),
        (DecisionTreeClassifier, {"max_depth": 5, "random_state": 42}),
        (
            RandomForestClassifier,
            {"max_depth": 5, "n_estimators": 10, "max_features": 1, "random_state": 42},
        ),
        (
            MLPClassifier,
            {
                # "alpha": 1,
                # "max_iter": 1000,
                "random_state": 42
            },
        ),
        (AdaBoostClassifier, {"random_state": 42}),
        (GaussianNB, {}),
    ]

    rmse_class_proba = [
        0.07,  # LogisticRegression
        0.16,  # KNeighborsClassifier
        0.1,  # GaussianProcessClassifier
        0.27,  # DecisionTreeClassifier
        0.11,  # RandomForestClassifier
        0.05,  # MLPClassifier
        0.20,  # AdaBoostClassifier
        0.07,  # GaussianNB
    ]

    rmse_class_sample = [
        0.08,  # LogisticRegression
        0.16,  # KNeighborsClassifier
        0.16,  # GaussianProcessClassifier
        0.15,  # DecisionTreeClassifier
        0.17,  # RandomForestClassifier
        0.11,  # MLPClassifier
        0.2,  # AdaBoostClassifier
        0.14,  # GaussianNB
    ]

    if XGB_AVAILABLE:
        probabilistic_classifiers.append((
            XGBClassifierModel,
            {
                "n_estimators": 1,
                "max_depth": 1,
                "max_leaves": 1,
                "random_state": 42,
            },
        ))
        rmse_class_proba.append(0.17)
        rmse_class_sample.append(0.18)

    if LGBM_AVAILABLE:
        probabilistic_classifiers.append((
            LightGBMClassifierModel,
            {
                "n_estimators": 1,
                "max_depth": 1,
                "num_leaves": 2,
                "verbosity": -1,
                "random_state": 42,
            },
        ))
        rmse_class_proba.append(0.06)
        rmse_class_sample.append(0.06)

    if CB_AVAILABLE:
        probabilistic_classifiers.append((
            CatBoostClassifierModel,
            {
                "iterations": 1,
                "depth": 1,
                "verbose": -1,
                "random_state": 42,
            },
        ))
        rmse_class_proba.append(0.15)
        rmse_class_sample.append(0.15)

    @pytest.mark.parametrize(
        "clf_params",
        process_model_list(probabilistic_classifiers),
    )
    def test_wrong_likelihood(self, clf_params):
        clf, kwargs = clf_params
        with pytest.raises(ValueError) as exc:
            _ = clf(lags=1, likelihood="does_not_exist", **kwargs)
        assert (
            str(exc.value)
            == "Invalid `likelihood='does_not_exist'`. Must be one of ['classprobability']"
        )

        with pytest.raises(ValueError) as exc:
            _ = _get_likelihood(
                likelihood="does_not_exist",
                n_outputs=1,
                available_likelihoods=[LikelihoodType.ClassProbability],
            )
        assert (
            str(exc.value)
            == "Invalid `likelihood='does_not_exist'`. Must be one of ['classprobability']"
        )

    @pytest.mark.parametrize(
        "clf_params",
        product(
            process_model_list(probabilistic_classifiers), [False, True], [False, True]
        ),
    )
    def test_class_proba_likelihood_median_pred_is_same_than_no_likelihood(
        self, clf_params
    ):
        # check that the model's prediction is the same with and without the likelihood
        # when predict_likelihood_parameters=False
        # Meaning _get_median_prediction on top of predict_proba produce the same output than the model predict
        (clf, kwargs), multi_series, multi_variate = clf_params

        _, series_train, _, _, labels_1, _, labels_2 = (
            generate_random_series_with_probabilities(
                multi_series=multi_series, multi_variate=multi_variate
            )
        )
        ocl = 5
        model = clf(lags=2, output_chunk_length=ocl, likelihood=None, **kwargs)
        assert model.likelihood is None
        model_likelihood = clf(lags=2, output_chunk_length=ocl, **kwargs)
        assert isinstance(model_likelihood.likelihood, ClassProbabilityLikelihood)

        model.fit(series_train)
        model_likelihood.fit(series_train)

        # native class prediction with maximum probability
        pred = model.predict(n=ocl, series=series_train)
        # likelihood class prediction with maximum probability
        pred_likelihood = model_likelihood.predict(
            n=ocl,
            series=series_train,
            predict_likelihood_parameters=False,
        )
        # likelihood class probability prediction
        pred_proba = model_likelihood.predict(
            n=ocl,
            series=series_train,
            predict_likelihood_parameters=True,
        )
        if not multi_series:
            pred = [pred]
            pred_likelihood = [pred_likelihood]
            pred_proba = [pred_proba]

        # compute class prediction with maximum probability
        pred_max_proba = []
        for idx, p in enumerate(pred_proba):
            p = p.values()
            # compute labels for each component
            p_1 = np.argmax(p[:, : len(labels_1)], axis=1, keepdims=True)
            if multi_variate:
                p_2 = np.argmax(p[:, len(labels_1) :], axis=1, keepdims=True)
                p_1 = np.concatenate([p_1, p_2], axis=1)
            pred_max_proba.append(pred[idx].with_values(p_1))

        # all predictions are the same
        assert pred_likelihood == pred
        assert pred_max_proba == pred

        with pytest.raises(ValueError) as err:
            model.predict(
                n=ocl + 1, series=series_train, predict_likelihood_parameters=True
            )
        assert (
            str(err.value) == "`predict_likelihood_parameters=True` is only"
            " supported for probabilistic models fitted with a likelihood."
        )

    @pytest.mark.parametrize(
        "clf_params",
        product(
            zip(process_model_list(probabilistic_classifiers), rmse_class_proba),
            [True, False],
            [True, False],
        ),
    )
    def test_class_probabilities_are_valid(self, clf_params):
        """
        Check class probabilities have correct shape and meaning in case of:
        - single series, univariate
        - single series, multivariate
        - multi series, univariate
        - multi series, multivariate
        Components of multi-variate do not have the same classes
        """

        ((clf, kwargs), rmse_margin), multi_series, multi_variate = clf_params
        model = clf(
            lags=2,
            **kwargs,
        )

        (
            series,
            series_train,
            series_test,
            probas_component1,
            labels_component1,
            probas_component2,
            labels_component2,
        ) = generate_random_series_with_probabilities(
            multi_series=multi_series, multi_variate=multi_variate
        )
        model.fit(series_train)

        # model_likelihood has ClassProbability likelihood
        list_of_probas = model.predict(
            n=1, series=series_test, predict_likelihood_parameters=True
        )

        list_of_probas2 = (
            model.untrained_model()
            .fit(series_train)
            .predict(n=1, series=series_test, predict_likelihood_parameters=True)
        )

        if not multi_series:
            series = [series]
            series_test = [series_test]
            list_of_probas = [list_of_probas]
            list_of_probas2 = [list_of_probas2]

        # Probability are reproducible
        for s, p1, p2 in zip(series, list_of_probas, list_of_probas2):
            vals1 = p1.all_values()
            vals2 = p2.all_values()
            np.testing.assert_allclose(vals1, vals2)

            # As many probability components as classes
            n_classes = len(labels_component1) + (
                len(labels_component2) if multi_variate else 0
            )
            assert p1.n_components == n_classes

            # Sum of class proba is 1 per component (x2 component must be 2)
            assert p1.values().sum() == pytest.approx(s.n_components)

            # component-specific probabilities sum to 1
            assert vals1[:, : len(labels_component1)].sum() == pytest.approx(1.0)
            if multi_variate:
                assert vals1[:, len(labels_component1) :].sum() == pytest.approx(1.0)

            # all labels are present in the class_labels
            assert np.all(model.class_labels[0] == labels_component1)
            if multi_variate:
                assert np.all(model.class_labels[1] == labels_component2)

        # verify that historical_forecasts returns on average approximately the same probabilities
        true_probas = (
            np.concatenate((probas_component1, probas_component2))
            if multi_variate
            else probas_component1
        )

        pred_probas = model.historical_forecasts(
            series=series,
            forecast_horizon=1,
            predict_likelihood_parameters=True,
            retrain=False,
            start=series_test[0].start_time() + 2 * series_test[0].freq,
            overlap_end=True,
        )
        avg_probas = np.array([pred.values().mean(axis=0) for pred in pred_probas])
        rmse = np.mean((avg_probas - true_probas) ** 2, axis=1) ** 0.5
        assert np.all(rmse <= rmse_margin)

    def test_bad_behavior_model_properties(self, caplog):
        class NoPredictProbaModel:
            def fit(*args):
                pass

            def predict(*args):
                pass

            def __sklearn_tags__(self):
                return Tags(estimator_type="classifier", target_tags=None)

        with caplog.at_level(logging.WARNING):
            SKLearnClassifierModel(model=NoPredictProbaModel(), lags=2)
            assert any([
                message.startswith(
                    "`model` has no method with name `predict_proba()`. "
                    "Probabilistic forecasting support deactivated."
                )
                for message in caplog.messages
            ])
        caplog.clear()

        with caplog.at_level(logging.WARNING):
            SKLearnClassifierModel(
                model=SVC(gamma=2, C=1, random_state=42, probability=True), lags=2
            )
            assert not any([
                message.startswith(
                    "`model` has no method with name `predict_proba()`. "
                )
                for message in caplog.messages
            ])

            # Without probability=True SVC has no "predict_proba" method
            SKLearnClassifierModel(model=SVC(gamma=2, C=1, random_state=42), lags=2)
            assert any(record.levelname == "WARNING" for record in caplog.records)
            assert any([
                message.startswith("`model` has no method with name `predict_proba()`.")
                for message in caplog.messages
            ])

        class NoClassLabelsModel:
            def fit(*args):
                pass

            def predict(*args):
                pass

            def predict_proba(*args):
                pass

            def __sklearn_tags__(self):
                return Tags(estimator_type="classifier", target_tags=None)

        model = NoClassLabelsModel()
        likelihood = ClassProbabilityLikelihood(n_outputs=1)
        with pytest.raises(ValueError) as err:
            likelihood.fit(model)
        assert (
            str(err.value)
            == "The model must have a `class_labels` attribute to fit the likelihood."
        )

        class NotSummingToOneProbasModel:
            def __init__(self, gap_to_one: int):
                self.gap_to_one = gap_to_one

            def fit(*args):
                pass

            def predict(*args):
                pass

            def predict_proba(self, *args):
                return np.array([[0.3, 0.2, 0.5], [0.3, 0.2, 0.5]]) - (
                    self.gap_to_one / 3
                )

            @property
            def class_labels(self):
                return [np.array([0, 1, 2])]

            def __sklearn_tags__(self):
                return Tags(estimator_type="classifier", target_tags=None)

        tolerance = 0.0000001

        # difference is smaller than tolerance, probability will be adjusted to sum up to one
        small_diff_model = NotSummingToOneProbasModel(tolerance / 2)
        small_diff_model.model = small_diff_model
        likelihood.fit(small_diff_model)
        likelihood.predict(
            small_diff_model,
            np.array([0]),
            num_samples=2,
            predict_likelihood_parameters=False,
        )

        # difference is larger than tolerance, error is raised
        large_diff_model = NotSummingToOneProbasModel(tolerance * 2)
        large_diff_model.model = large_diff_model
        likelihood.fit(large_diff_model)
        with pytest.raises(ValueError) as err:
            likelihood.predict(
                large_diff_model,
                np.array([0]),
                num_samples=2,
                predict_likelihood_parameters=False,
            )
        assert (
            str(err.value)
            == "The class probabilities returned by the model do not sum to one"
        )

    def test_class_probability_component_names(self):
        _, series_train, _, _, _, _, _ = generate_random_series_with_probabilities(
            multi_series=False, multi_variate=True
        )

        component_names = [
            "component1_p0",
            "component1_p1",
            "component1_p2",
            "component2_p0",
            "component2_p1",
        ]

        model = SKLearnClassifierModel(lags=1, output_chunk_length=2)

        # component_names before fit throws an error
        with pytest.raises(ValueError) as err:
            model.likelihood.component_names(series_train)
        assert str(err.value) == "The likelihood has not been fitted yet."

        # once fitted, component_names are correct
        model.fit(series_train)
        assert model.likelihood.component_names(series=series_train) == component_names
        assert (
            model.likelihood.component_names(components=series_train.components)
            == component_names
        )

        # predicted component names are correct
        preds = model.predict(n=2, predict_likelihood_parameters=True)

        assert np.all(preds.components == component_names)

        with pytest.raises(ValueError) as err:
            _ = model.likelihood.component_names(
                series=series_train, components=series_train.components
            )
        assert (
            str(err.value) == "Only one of `series` or `components` must be specified."
        )

    @pytest.mark.parametrize(
        "clf_params",
        product(
            zip(process_model_list(probabilistic_classifiers), rmse_class_sample),
            [False, True],
        ),
    )
    def test_multi_sample_with_class_probabilities(self, clf_params):
        """
        The distribution of samples corresponds to the distribution of labels in the TS
        """
        ((clf, kwargs), model_rmse), multi_variate = clf_params
        model = clf(lags=2, **kwargs)

        (
            _,
            series_train,
            _,
            probas_component1,
            labels_component1,
            probas_component2,
            labels_component2,
        ) = generate_random_series_with_probabilities(
            multi_series=False, multi_variate=multi_variate
        )

        model.fit(series_train)

        # predict_likelihood_parameters is not supported with multiple samples
        with pytest.raises(ValueError) as err:
            model.predict(n=1, num_samples=2, predict_likelihood_parameters=True)
        assert (
            str(err.value)
            == "`predict_likelihood_parameters=True` is only supported for `num_samples=1`, received 2."
        )

        # predict with multiple samples and verify the label probability
        prediction = model.predict(n=5, num_samples=1000)
        preds_c = [prediction.all_values()[:, 0]]
        probas_c = [probas_component1]
        labels_c = [labels_component1]
        if multi_variate:
            preds_c.append(prediction.all_values()[:, 1])
            probas_c.append(probas_component2)
            labels_c.append(labels_component2)

        for idx, (preds, probas, labels) in enumerate(zip(preds_c, probas_c, labels_c)):
            pred_probas = pd.value_counts(preds.flatten(), normalize=True).to_dict()
            pred_probas = [pred_probas[label] for label in labels]
            rmse = np.mean((pred_probas - probas) ** 2) ** 0.5
            assert rmse < model_rmse

        # reproducible samples when call order is the same (and also with transferable series)
        model = model.untrained_model().fit(series_train)
        preds_2 = model.predict(n=5, num_samples=1000, series=series_train)
        assert preds_2 == prediction

        # different samples when call order changes
        preds_3 = model.predict(n=5, num_samples=1000)
        assert preds_3 != prediction

        # same samples when random_state is set
        preds_4 = model.predict(
            n=5, num_samples=1000, series=series_train, random_state=42
        )
        preds_5 = model.predict(
            n=5, num_samples=1000, series=series_train, random_state=42
        )
        assert preds_4 == preds_5

    @pytest.mark.parametrize(
        "params",
        product(
            [True, False],  # multi_variate
            [True, False],  # multi_model
        ),
    )
    def test_historical_forecast(self, params):
        multi_model, multi_variate = params

        _, series_train, _, _, _, _, _ = generate_random_series_with_probabilities(
            multi_series=False, multi_variate=multi_variate
        )

        _, series_multi_variate, _, _, _, _, _ = (
            generate_random_series_with_probabilities(
                multi_series=False, multi_variate=True
            )
        )

        # without covariate
        model = SKLearnClassifierModel(
            model=LogisticRegression(), lags=4, multi_models=multi_model
        )
        result = model.historical_forecasts(
            series=series_train,
            future_covariates=None,
            start=0.8,
            forecast_horizon=1,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False,
        )
        assert len(result) == 21

        # With past covariate
        model = SKLearnClassifierModel(
            model=LogisticRegression(),
            lags=5,
            lags_past_covariates=5,
            multi_models=multi_model,
        )
        result = model.historical_forecasts(
            series=series_train,
            past_covariates=series_multi_variate,
            start=0.8,
            forecast_horizon=1,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False,
        )
        assert len(result) == 21

        # with past covariate and output_chunk_length
        model = SKLearnClassifierModel(
            model=LogisticRegression(),
            lags=5,
            lags_past_covariates=5,
            output_chunk_length=5,
            multi_models=multi_model,
        )
        result = model.historical_forecasts(
            series=series_train,
            past_covariates=series_multi_variate,
            start=0.8,
            forecast_horizon=1,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False,
        )
        assert len(result) == 21

        # forecast_horizon > output_chunk_length
        model = SKLearnClassifierModel(
            model=LogisticRegression(),
            lags=5,
            lags_past_covariates=5,
            output_chunk_length=1,
            multi_models=multi_model,
        )
        result = model.historical_forecasts(
            series=series_train,
            past_covariates=series_multi_variate,
            start=0.8,
            forecast_horizon=3,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False,
        )
        assert len(result) == 19

        # Likelihood parameters
        model = SKLearnClassifierModel(
            model=LogisticRegression(),
            lags=5,
            lags_past_covariates=5,
            multi_models=multi_model,
        )
        result = model.historical_forecasts(
            series=series_train,
            past_covariates=series_multi_variate,
            start=0.8,
            forecast_horizon=1,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False,
            predict_likelihood_parameters=True,
        )
        assert len(result) == 21

        # Sampled prediction
        model = SKLearnClassifierModel(
            model=LogisticRegression(),
            lags=5,
            lags_past_covariates=5,
            multi_models=multi_model,
        )
        result = model.historical_forecasts(
            series=series_train,
            past_covariates=series_multi_variate,
            start=0.8,
            forecast_horizon=1,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False,
            num_samples=100,
        )
        assert len(result) == 21



import numpy as np
import pytest

from darts import TimeSeries
from darts.utils.likelihood_models.base import Likelihood, LikelihoodType


class TestLikelihoodModel:
    def test_likelihood_component_names(self):
        lkl = Likelihood(
            likelihood_type=LikelihoodType.Gaussian, parameter_names=["mu", "sigma"]
        )

        components = ["a", "b"]
        series = TimeSeries.from_values(
            values=np.zeros((3, len(components))), columns=components
        )
        # cannot give series and components at the same time
        with pytest.raises(ValueError):
            _ = lkl.component_names(series=series, components=components)

        names_expected = ["a_mu", "a_sigma", "b_mu", "b_sigma"]
        names_1 = lkl.component_names(series=series)
        names_2 = lkl.component_names(components=components)
        assert names_1 == names_2 == names_expected



import itertools

import numpy as np
import pytest
import sklearn.metrics as sklearn_metrics

import darts.metrics.metrics as metrics
import darts.metrics.utils as utils
from darts import TimeSeries


class TestClassificationMetrics:
    np.random.seed(42)

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [0, 1, 2],  # deterministic labels, class probabilities, sampled labels
            [None, 4, [4, 1], [1, 2, 3, 4]],  # selected labels
            ["weighted", "micro", "macro", None],  # label reduction
            [
                (
                    metrics.accuracy,
                    sklearn_metrics.accuracy_score,
                    np.nanmean,
                ),
                (
                    metrics.recall,
                    sklearn_metrics.recall_score,
                    np.nanmean,
                ),
                (
                    metrics.precision,
                    sklearn_metrics.precision_score,
                    np.nanmean,
                ),
                (
                    metrics.f1,
                    sklearn_metrics.f1_score,
                    np.nanmean,
                ),
                (
                    metrics.confusion_matrix,
                    sklearn_metrics.confusion_matrix,
                    np.nansum,
                ),
            ],
        ),
    )
    def test_classification_metric(self, config):
        method, labels, label_reduction, (metric, skl_metric, comp_reduction) = config
        comp1_labels = np.array([2, 4])
        comp1_probas = np.array([
            [0.1, 0.9],
            [0.4, 0.6],
            [0.9, 0.1],
            [0.7, 0.3],
        ])
        comp2_labels = np.array([1, 3, 4])
        comp2_probas = np.array([
            [0.3, 0.2, 0.5],
            [0.6, 0.1, 0.3],
            [0.2, 0.6, 0.2],
            [0.3, 0.1, 0.6],
        ])

        if metric == metrics.accuracy:
            # accuracy is not label specific
            kwargs = {}
            skl_kwargs = {}
        else:
            kwargs = {"labels": labels}
            labels = [labels] if isinstance(labels, int) else labels
            if metric == metrics.confusion_matrix:
                skl_kwargs = {"labels": labels if labels is not None else [1, 2, 3, 4]}
            else:
                kwargs["label_reduction"] = label_reduction
                skl_kwargs = {
                    "average": label_reduction,
                    "zero_division": np.nan,
                    "labels": labels
                    if not (labels is None and label_reduction is None)
                    else [1, 2, 3, 4],
                }

        if method == 0:
            # deterministic class labels
            names = ["comp1", "comp2"]
            vals = np.concatenate(
                [
                    comp1_labels[np.argmax(comp1_probas, axis=1), None],
                    comp2_labels[np.argmax(comp2_probas, axis=1), None],
                ],
                axis=1,
            )
        elif method == 1:
            # class probabilities
            names = ["comp1_p2", "comp1_p4", "comp2_p1", "comp2_p3", "comp2_p4"]
            vals = np.concatenate([comp1_probas, comp2_probas], axis=1)
        else:
            # sampled class labels
            names = ["comp1", "comp2"]
            num_samples = 1000
            vals = np.zeros((4, 2, num_samples))
            for idx in range(len(comp1_probas)):
                vals[idx, 0, :] = np.random.choice(
                    comp1_labels, size=num_samples, p=comp1_probas[idx]
                )
                vals[idx, 1, :] = np.random.choice(
                    comp2_labels, size=num_samples, p=comp2_probas[idx]
                )

        # predicted labels are comp1: [4, 4, 2, 2], comp2: [4, 1, 3, 4]
        y_pred = TimeSeries.from_values(vals, names)
        y_true = TimeSeries.from_values(
            np.array([
                [4, 4, 2, 2],
                [4, 1, 4, 1],
            ]).T[:, :, np.newaxis],
            ["comp1", "comp2"],
        )

        # get predicted labels with highest probability / frequency
        y_true_vals, y_pred_vals = metrics._get_values_or_raise(
            y_true, y_pred, True, remove_nan_union=False, is_classification=True
        )

        # compare with sklearn metrics
        scores_expected = np.array([
            skl_metric(y_true_vals[:, i], y_pred_vals[:, i], **skl_kwargs)
            for i in range(y_true_vals.shape[1])
        ])
        scores_expected_ = scores_expected
        if label_reduction is None and metric not in [
            metrics.confusion_matrix,
            metrics.accuracy,
        ]:
            # concatenate if reduction is None (e.g. Darts gives label-specific scores in component dim, similar to
            # quantile metrics)
            scores_expected_ = np.concatenate(scores_expected)

        # without component reduction
        scores_actual = metric(
            y_true,
            y_pred,
            component_reduction=None,
            **kwargs,
        )

        if metric == metrics.confusion_matrix:
            # replace NaNs with 0.0 to compare with sklearn metrics (Darts puts np.nan into missing labels,
            # sklearn puts zeros)
            scores_actual[np.isnan(scores_actual)] = 0.0

        np.testing.assert_array_almost_equal(
            scores_actual,
            scores_expected_,
        )

        # with mean component reduction
        np.testing.assert_array_almost_equal(
            metric(
                y_true,
                y_pred,
                **kwargs,
            ),
            comp_reduction(scores_expected, axis=0),
        )

    def test_wrong_pred_component_names(self):
        y_pred = TimeSeries.from_values(
            values=np.array([[0.1, 0.9, 0.3, 0.7]]),
            columns=["comp0_p0", "comp0_p4", "comp1_p0", "comp1_p3"],
        )
        y_true = TimeSeries.from_values(
            values=np.array([[4, 3]]),
            columns=["comp0", "comp1"],
        )
        assert metrics.accuracy(y_true, y_pred) == 1.0

        # pred components should have names ["comp0*", "comp1*"]
        y_pred._components = y_pred._components.str.replace("comp0", "comp3")
        with pytest.raises(ValueError) as err:
            metrics.accuracy(y_true, y_pred)
        assert str(err.value).startswith(
            "Could not resolve the predicted components for the classification metric"
        )

    def test_wrong_number_of_pred_components(self):
        y_pred = TimeSeries.from_values(
            values=np.array([[0.1, 0.9, 0.3, 0.7]]),
            columns=["comp1_p0", "comp1_p1", "comp2_p0", "comp2_p1"],
        )
        y_true = TimeSeries.from_values(
            values=np.array([[4, 4, 4]]),
            columns=["comp0", "comp1", "comp2"],
        )

        # pred should have 3 components
        with pytest.raises(ValueError) as err:
            metrics.accuracy(y_true, y_pred)
        assert str(err.value).startswith(
            "Could not resolve the predicted components for the classification metric"
        )

    def test_non_integer_label(self):
        y_pred = TimeSeries.from_values(
            values=np.array([[0.1, 0.9]]),
            columns=["comp0_p2.5", "comp0_p4"],
        )
        y_true = TimeSeries.from_values(
            values=np.array([[4]]),
            columns=["comp0"],
        )

        # pred should have 3 components
        with pytest.raises(ValueError) as err:
            metrics.accuracy(y_true, y_pred)
        assert str(err.value).startswith(
            "Could not parse class label from name: comp0_p2.5"
        )

    def test_mode(self):
        # Test with a simple list of integers
        data = np.array([2, 1, 2, 3, 4])
        assert utils._mode(data) == 2

        # Test with a list of strings
        data = np.array(["apple", "banana", "apple", "orange"])
        assert utils._mode(data) == "apple"

        # Test with a list where all elements are unique; gives first element with highest count
        data = np.array([1, 2, 3, 4])
        assert utils._mode(data) == 1

    def test_get_highest_count_label_single_step_single_component(self):
        # Test with a simple list of integers
        data = np.array([2, 1, 2, 3, 4]).reshape((1, 1, 5))
        res = utils._get_highest_count_label(data)
        assert res.shape == (1, 1, 1)
        assert np.all(res == 2)

        # Test with a list of integers with NaN values
        data = np.array([2, 1, 2, np.nan, 4]).reshape((1, 1, 5))
        res = utils._get_highest_count_label(data)
        assert res.shape == (1, 1, 1)
        assert np.all(res == 2)

        # Test with a list of integers where NaN values have highest count
        data = np.array([2, 1, np.nan, np.nan, 4]).reshape((1, 1, 5))
        res = utils._get_highest_count_label(data)
        assert res.shape == (1, 1, 1)
        assert np.isnan(res)

        # Test with a list of strings
        data = np.array(["apple", "banana", "apple", "orange"]).reshape((1, 1, 4))
        res = utils._get_highest_count_label(data)
        assert res.shape == (1, 1, 1)
        assert np.all(res == "apple")

        # Test with a list where all elements are unique; gives lowest value element with highest count
        data = np.array([4, 1, 3, 2]).reshape((1, 1, 4))
        res = utils._get_highest_count_label(data)
        assert res.shape == (1, 1, 1)
        assert np.all(res == 1)

    def test_get_highest_count_label_multi_step_multi_component(self):
        # Test with a simple list of integers
        data = np.array([
            [[2, 1, 2, 3, 4], [1, 2, 3, 3, 5]],
            [[1, 2, 3, 4, 5], [2, 5, 5, 4, 4]],
            [[-1, 0, 0, -1, 5], [0, 0, 0, 0, 0]],
        ])
        res = utils._get_highest_count_label(data)
        assert res.shape == (3, 2, 1)
        assert np.all(
            res
            == np.array([
                [[2], [3]],
                [[1], [4]],
                [[-1], [0]],
            ])
        )

    def test_get_highest_probability_label_single_step_single_component(self):
        # Test with a binary labels
        data = np.array([0.1, 0.9]).reshape((1, 2, 1))
        res = utils._get_highest_probability_label(data, ["a_p0", "a_p1"], ["a"])
        assert res.shape == (1, 1, 1)
        assert np.all(res == 1)

        # Test with three labels and unsorted label names
        data = np.array([0.1, 0.8, 0.1]).reshape((1, 3, 1))
        res = utils._get_highest_probability_label(
            data, ["a_p0", "a_p2", "a_p1"], ["a"]
        )
        assert res.shape == (1, 1, 1)
        assert np.all(res == 2)

        # Test with three labels where two labels have the highest probability (returns the first one)
        data = np.array([0.4, 0.2, 0.4]).reshape((1, 3, 1))
        res = utils._get_highest_probability_label(
            data, ["a_p3", "a_p2", "a_p1"], ["a"]
        )
        assert res.shape == (1, 1, 1)
        assert np.all(res == 3)

        with pytest.raises(ValueError) as err:
            # label cannot be converted to integer
            utils._get_highest_probability_label(
                data, ["a_p3", "a_p2.0", "a_p1"], ["a"]
            )
        assert str(err.value).startswith(
            "Could not parse class label from name: a_p2.0"
        )

    def test_get_highest_probability_label_multi_step_multi_component(self):
        # Test with a binary labels
        data = np.array([
            [[0.1], [0.9], [0.6], [0.2], [0.2]],
            [[0.3], [0.7], [0.4], [0.5], [0.1]],
            [[0.5], [0.5], [0.2], [0.0], [0.8]],
        ])
        res = utils._get_highest_probability_label(
            data, ["a_p0", "a_p1", "b_p5", "b_p4", "b_p1"], ["a", "b"]
        )
        assert res.shape == (3, 2, 1)
        assert np.all(
            res
            == np.array([
                [[1], [5]],
                [[1], [4]],
                [[0], [1]],
            ])
        )

    def test_wrong_label_reduction(self):
        y = TimeSeries.from_values(np.array([[4]]))

        # invalid label reduction method
        with pytest.raises(ValueError) as err:
            metrics.recall(y, y, label_reduction="invalid")
        assert str(err.value).startswith("Invalid `label_reduction` value: `invalid`")
